import logging
import os
import time
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Callable, Dict, Iterable, List, Set

from opensearchpy import OpenSearch
from opensearchpy.exceptions import (
    ConnectionError as OpenSearchConnectionError,
    ConnectionTimeout,
    RequestError,
    TransportError,
)
from pydantic import validator
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from acryl_datahub_cloud.elasticsearch.config import ElasticSearchClientConfig
from acryl_datahub_cloud.elasticsearch.graph_service import ElasticGraphRow
from datahub.configuration import ConfigModel
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source_report.ingestion_stage import IngestionStageReport
from datahub.metadata.schema_classes import AuditStampClass, LineageFeaturesClass

logger = logging.getLogger(__name__)

SYSTEM_ACTOR = "urn:li:corpuser:__datahub_system"


class LineageFeaturesSourceConfig(ConfigModel):
    enabled: bool = True
    materialize_entities: bool = False
    search_index: ElasticSearchClientConfig = ElasticSearchClientConfig()
    query_timeout: int = 30
    extract_batch_size: int = 3000
    max_retries: int = 3
    retry_delay_seconds: int = 5
    retry_backoff_multiplier: float = 2.0

    @validator("max_retries")
    def validate_max_retries(cls, v: int) -> int:
        if v < 1:
            raise ValueError("max_retries must be at least 1")
        return v

    @validator("retry_delay_seconds")
    def validate_retry_delay_seconds(cls, v: int) -> int:
        if v < 1:
            raise ValueError("retry_delay_seconds must be at least 1")
        return v

    @validator("retry_backoff_multiplier")
    def validate_retry_backoff_multiplier(cls, v: float) -> float:
        if v < 1.0:
            raise ValueError("retry_backoff_multiplier must be at least 1.0")
        return v


@dataclass
class LineageExtractGraphSourceReport(SourceReport, IngestionStageReport):
    valid_urns_count: int = 0
    upstream_count: int = 0
    downstream_count: int = 0
    edges_scanned: int = 0
    skipped_materialized_urns_count: int = 0


@platform_name(id="datahub", platform_name="DataHub")
@config_class(LineageFeaturesSourceConfig)
@support_status(SupportStatus.INCUBATING)
class DataHubLineageFeaturesSource(Source):
    """
    DataHub Lineage Features Source that extracts lineage information from Elasticsearch/OpenSearch.
    """

    platform = "datahub"

    def __init__(
        self, config: LineageFeaturesSourceConfig, ctx: PipelineContext
    ) -> None:
        super().__init__(ctx)
        self.config: LineageFeaturesSourceConfig = config
        self.report = LineageExtractGraphSourceReport()
        self.opened_files: List[str] = []

        self.valid_urns: Set[str] = set()
        self.upstream_counts: Dict[str, int] = defaultdict(int)
        self.downstream_counts: Dict[str, int] = defaultdict(int)
        self.last_print_time = time.time()

    def _get_retry_decorator(
        self,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """Create a retry decorator based on config parameters"""

        def should_retry_exception(exception: Exception) -> bool:
            """Custom retry predicate for OpenSearch exceptions"""
            if isinstance(
                exception,
                (
                    OpenSearchConnectionError,
                    ConnectionTimeout,
                    RequestError,
                    TransportError,
                ),
            ):
                return True
            # Also retry on general connection and timeout errors
            if isinstance(exception, (ConnectionError, TimeoutError)):
                return True
            return False

        return retry(
            retry=retry_if_exception_type(
                (
                    OpenSearchConnectionError,
                    ConnectionTimeout,
                    RequestError,
                    TransportError,
                    ConnectionError,
                    TimeoutError,
                )
            ),
            stop=stop_after_attempt(self.config.max_retries),
            wait=wait_exponential(
                multiplier=self.config.retry_backoff_multiplier,
                min=self.config.retry_delay_seconds,
                max=30,
            ),
            before_sleep=before_sleep_log(logger, logging.WARNING),
            reraise=True,
        )

    def _create_pit_with_retry(self, server: OpenSearch, index: str) -> str:
        """Create a Point-in-Time (PIT) with retry logic"""
        retry_decorator = self._get_retry_decorator()

        @retry_decorator
        def _create_pit() -> str:
            logger.debug(f"Creating PIT for index: {index}")
            response = server.create_pit(index, keep_alive="10m")
            pit = response.get("pit_id")
            if not pit:
                raise Exception("Failed to create PIT - no pit_id returned")
            logger.debug(f"Successfully created PIT: {pit}")
            return pit

        return _create_pit()

    def _search_with_retry(
        self, server: OpenSearch, query: dict, batch_size: int
    ) -> dict:
        """Execute search with retry logic"""
        retry_decorator = self._get_retry_decorator()

        @retry_decorator
        def _search() -> dict:
            logger.debug(f"Executing search with batch size: {batch_size}")
            return server.search(
                body=query,
                size=batch_size,
                params={"timeout": self.config.query_timeout},
            )

        return _search()

    def _delete_pit_with_retry(self, server: OpenSearch, pit: str) -> None:
        """Delete Point-in-Time (PIT) with retry logic"""
        retry_decorator = self._get_retry_decorator()

        @retry_decorator
        def _delete_pit() -> None:
            logger.debug(f"Deleting PIT: {pit}")
            server.delete_pit(body={"pit_id": pit})
            logger.debug(f"Successfully deleted PIT: {pit}")

        _delete_pit()

    def _create_opensearch_client_with_retry(self) -> OpenSearch:
        """Create OpenSearch client with retry logic"""
        retry_decorator = self._get_retry_decorator()

        @retry_decorator
        def _create_client() -> OpenSearch:
            logger.debug(
                f"Creating OpenSearch client for endpoint: {self.config.search_index.endpoint}"
            )
            return OpenSearch(
                [self.config.search_index.endpoint],
                http_auth=(
                    self.config.search_index.username,
                    self.config.search_index.password,
                ),
                use_ssl=self.config.search_index.use_ssl,
            )

        return _create_client()

    def _update_report(self) -> None:
        """
        Information to see whether we are close to hitting the memory limits
        """
        self.report.valid_urns_count = len(self.valid_urns)
        self.report.upstream_count = len(self.upstream_counts.keys())
        self.report.downstream_count = len(self.downstream_counts.keys())

    def _print_report(self) -> None:
        """
        Printing is required like this because the report is only printed
        when the workunits are yielded
        In case of background processes we won't know the progress if this is not done
        """
        time_taken = round(time.time() - self.last_print_time, 1)
        # Print report every 2 minutes
        if time_taken > 120:
            self._update_report()
            self.last_print_time = time.time()
            logger.info(f"\n{self.report.as_string()}")

    def process_batch(self, results: Iterable[dict]) -> None:
        for doc in results:
            self._print_report()
            row = ElasticGraphRow.from_elastic_doc(doc["_source"])
            self.report.edges_scanned += 1
            if (
                row.source_urn in self.valid_urns
                and row.destination_urn in self.valid_urns
            ):
                self.upstream_counts[row.source_urn] += 1
                self.downstream_counts[row.destination_urn] += 1

    def populate_valid_urns(self) -> None:
        graph = self.ctx.require_graph("Load non soft-deleted urns")
        for urn in graph.get_urns_by_filter(batch_size=self.config.extract_batch_size):
            self._print_report()
            self.valid_urns.add(urn)

    def get_workunits(self) -> Iterable[MetadataWorkUnit]:
        with self.report.new_stage("Load valid URNs"):
            self.populate_valid_urns()

        timestamp = datetime.now(tz=timezone.utc)
        server = self._create_opensearch_client_with_retry()

        query = {
            "query": {
                "bool": {
                    "should": [
                        {"term": {"relationshipType": "Consumes"}},
                        {"term": {"relationshipType": "DownstreamOf"}},
                        {"term": {"relationshipType": "TrainedBy"}},
                        {"term": {"relationshipType": "UsedBy"}},
                        {"term": {"relationshipType": "MemberOf"}},
                        {"term": {"relationshipType": "DerivedFrom"}},
                        {"term": {"relationshipType": "Produces"}},
                        {"term": {"relationshipType": "DashboardContainsDashboard"}},
                        {
                            "bool": {
                                "must": [
                                    {"term": {"relationshipType": "Contains"}},
                                    {"term": {"source.entityType": "dashboard"}},
                                    {"term": {"destination.entityType": "chart"}},
                                ]
                            }
                        },
                    ],
                },
            },
            "sort": [
                {"source.urn": {"order": "desc"}},
                {"destination.urn": {"order": "desc"}},
                {"relationshipType": {"order": "desc"}},
                {"lifecycleOwner": {"order": "desc"}},
            ],
        }

        index = f"{self.config.search_index.index_prefix}graph_service_v1"
        pit = self._create_pit_with_retry(server, index)

        # TODO: Save PIT, we can resume processing based on <pit, search_after> tuple
        query.update({"pit": {"id": pit, "keep_alive": "10m"}})

        # TODO: Using slicing we can parallelize the ES calls below:
        # https://opensearch.org/docs/latest/search-plugins/searching-data/point-in-time/#search-slicing
        batch_size = self.config.extract_batch_size
        with self.report.new_stage("Extract lineage features"):
            try:
                while True:
                    results = self._search_with_retry(server, query, batch_size)
                    self.process_batch(results["hits"]["hits"])
                    if len(results["hits"]["hits"]) < batch_size:
                        break
                    query.update({"search_after": results["hits"]["hits"][-1]["sort"]})
            except Exception as e:
                logger.error(f"Error during lineage extraction: {e}")
                self.report.report_failure(
                    title="Lineage extraction failed",
                    message="Failed to extract lineage features from Elasticsearch",
                    context=f"Error: {str(e)}",
                    exc=e,
                )
                # Ensure PIT is cleaned up even on error
                try:
                    self._delete_pit_with_retry(server, pit)
                except Exception as cleanup_error:
                    logger.warning(
                        f"Failed to cleanup PIT after error: {cleanup_error}"
                    )
                raise
        self._update_report()
        self._delete_pit_with_retry(server, pit)

        self.report.new_stage("start emission of lineage features")
        # In Python 3.9, can be replaced by `self.self.upstream_counts.keys() | self.downstream_counts.keys()`
        for urn in set(self.upstream_counts.keys()).union(
            self.downstream_counts.keys()
        ):
            if (not self.config.materialize_entities) and urn not in self.valid_urns:
                self.report.skipped_materialized_urns_count += 1
                continue
            logger.debug(
                f"{urn}: {self.upstream_counts[urn]}, {self.downstream_counts[urn]}"
            )
            wu = MetadataChangeProposalWrapper(
                entityUrn=urn,
                aspect=LineageFeaturesClass(
                    upstreamCount=self.upstream_counts[urn],
                    downstreamCount=self.downstream_counts[urn],
                    computedAt=AuditStampClass(
                        time=int(timestamp.timestamp() * 1000),
                        actor=SYSTEM_ACTOR,
                    ),
                ),
            ).as_workunit()
            self.report.report_workunit(wu)
            yield wu
        # So previous stage's calculations are done
        self.report.new_stage("end emission of lineage features")

    def get_report(self) -> SourceReport:
        return self.report

    def close(self) -> None:
        for file in self.opened_files:
            os.remove(file)
        return super().close()



import json
import logging
from datetime import date, datetime, timezone
from enum import Enum
from typing import Any, Callable, Dict, Iterable, List, Optional

import pandas as pd
from pydantic import BaseModel

from acryl_datahub_cloud.elasticsearch.graph_service import BaseModelRow
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.graph.client import DataHubGraph
from datahub.ingestion.graph.filters import RawSearchFilter
from datahub.metadata.schema_classes import (
    DomainPropertiesClass,
    FormAssociationClass,
    FormInfoClass,
    FormsClass,
    FormStateClass,
    FormTypeClass,
)

logger = logging.getLogger(__name__)


class FormStatus(str, Enum):
    NOT_STARTED = "not_started"
    IN_PROGRESS = "in_progress"
    COMPLETED = "complete"


class QuestionStatus(str, Enum):
    NOT_STARTED = "not_started"
    COMPLETED = "complete"


class FormType(str, Enum):
    DOCUMENTATION = "documentation"
    VERIFICATION = "verification"


class FormReportingRow(BaseModelRow):
    form_urn: str
    form_type: FormType
    form_assigned_date: Optional[date]
    form_completed_date: Optional[date]
    form_status: FormStatus
    question_id: str
    question_status: QuestionStatus
    question_completed_date: Optional[date]
    assignee_urn: Optional[str]
    asset_urn: str
    platform_urn: str
    platform_instance_urn: Optional[str]
    domain_urn: Optional[str]
    parent_domain_urn: Optional[str]
    asset_verified: Optional[bool]
    snapshot_date: date


class FormData:
    def get_data(
        self,
        on_asset_scanned: Optional[Callable[[str], Any]] = None,
        on_form_scanned: Optional[Callable[[str], Any]] = None,
    ) -> Iterable[FormReportingRow]:
        raise NotImplementedError

    def to_mcps(self) -> Iterable[MetadataChangeProposalWrapper]:
        raise NotImplementedError


class FormRegistry:
    def __init__(self, graph: DataHubGraph):
        self.registry: Dict[str, Optional[FormInfoClass]] = {}
        self.graph = graph

    def get_form(self, form_id: str) -> Optional[FormInfoClass]:
        if form_id not in self.registry:
            self._load_form(form_id)
        return self.registry[form_id]

    def _load_form(self, form_id: str) -> None:
        form_urn = (
            f"urn:li:form:{form_id}"
            if not form_id.startswith("urn:li:form:")
            else form_id
        )

        form_info = self.graph.get_aspect(form_urn, FormInfoClass)
        self.registry[form_id] = form_info


class DomainRegistry:
    def __init__(self, graph: DataHubGraph):
        self.registry: Dict[str, Optional[DomainPropertiesClass]] = {}
        self.graph = graph

    def get_domain(self, domain_id: str) -> Optional[DomainPropertiesClass]:
        if domain_id not in self.registry:
            self._load_domain(domain_id)
        return self.registry[domain_id]

    def _load_domain(self, domain_id: str) -> None:
        domain_urn = (
            f"urn:li:domain:{domain_id}"
            if not domain_id.startswith("urn:li:domain:")
            else domain_id
        )
        domain_props = self.graph.get_aspect(domain_urn, DomainPropertiesClass)
        self.registry[domain_id] = domain_props


class DataHubFormReportingData(FormData):
    class DataHubDatasetSearchRow(BaseModel):
        urn: str
        owners: List[str] = []
        completedForms: List[str] = []
        incompleteForms: List[str] = []
        verifiedForms: List[str] = []
        completedFormsIncompletePromptIds: List[str] = []
        completedFormsIncompletePromptResponseTimes: List[str] = []
        completedFormsCompletedPromptIds: List[str] = []
        completedFormsCompletedPromptResponseTimes: List[str] = []
        incompleteFormsIncompletePromptIds: List[str] = []
        incompleteFormsIncompletePromptResponseTimes: List[str] = []
        incompleteFormsCompletedPromptIds: List[str] = []
        incompleteFormsCompletedPromptResponseTimes: List[str] = []
        platform: str = ""
        platformInstance: Optional[str] = None
        domains: List[str] = []

    def __init__(self, graph: DataHubGraph, allowed_forms: Optional[List[str]] = None):
        self.graph: DataHubGraph = graph
        self.form_registry = FormRegistry(graph)
        self.domain_registry = DomainRegistry(graph)
        self.snapshot_date = datetime.now().date()
        self.allowed_forms = allowed_forms

    def get_dataframe(
        self,
        on_asset_scanned: Callable[[str], Any],
        on_form_scanned: Callable[[str], Any],
    ) -> pd.DataFrame:
        return pd.DataFrame(
            x.dict()
            for x in self.get_data(
                on_asset_scanned=on_asset_scanned, on_form_scanned=on_form_scanned
            )
        )

    def get_form_existence_or_filters(self) -> RawSearchFilter:
        """
        Datasets must either have completedForms or incompleteForms assigned to
        them
        """
        if self.allowed_forms:
            return [
                {
                    "and": [
                        {
                            "field": "completedForms",
                            "condition": "EQUAL",
                            "values": self.allowed_forms,
                        }
                    ]
                },
                {
                    "and": [
                        {
                            "field": "incompleteForms",
                            "condition": "EQUAL",
                            "values": self.allowed_forms,
                        }
                    ]
                },
            ]
        else:
            return [
                {
                    "and": [
                        {
                            "field": "completedForms",
                            "condition": "EXISTS",
                        }
                    ]
                },
                {
                    "and": [
                        {
                            "field": "incompleteForms",
                            "condition": "EXISTS",
                        }
                    ]
                },
            ]

    def is_published(self, form_urn: str) -> bool:
        form_info = self.form_registry.get_form(form_urn)
        return (
            form_info.status.state == FormStateClass.PUBLISHED if form_info else False
        )

    def form_published_time(self, form_urn: str) -> float:
        form_info = self.form_registry.get_form(form_urn)
        is_published = (
            form_info.status.state == FormStateClass.PUBLISHED if form_info else False
        )
        return (
            form_info.status.lastModified.time / 1000
            if form_info and form_info.status.lastModified and is_published
            else 0
        )

    def assigned_to_asset_time(self, form_association: FormAssociationClass) -> float:
        return form_association.created.time / 1000 if form_association.created else 0

    def assignment_time(self, form_association: FormAssociationClass) -> float:
        published_time = self.form_published_time(form_association.urn)
        assigned_to_asset_time = self.assigned_to_asset_time(form_association)
        return max(published_time, assigned_to_asset_time)

    # For a given asset, the assigned date is the more recent of the published date and the time this was actually assigned.
    # Assets can be assigned before publishing, but we don't want to show that date because it's not open to the public yet.
    # Assets can also be assigned after publishing, so we should show that date for those assets.
    def form_assigned_date(
        self, search_row: DataHubDatasetSearchRow
    ) -> Dict[str, Optional[date]]:
        form_assigned_dates: Dict[str, Optional[date]] = {}
        forms = self.graph.get_aspect(search_row.urn, FormsClass)
        if not forms:
            return form_assigned_dates
        assert forms, f"Forms aspect not found for {search_row.urn}"
        for incomplete_form in forms.incompleteForms:
            is_published = self.is_published(incomplete_form.urn)
            assignment_time = self.assignment_time(incomplete_form)
            form_assigned_dates[incomplete_form.urn] = (
                datetime.fromtimestamp(
                    assignment_time,
                    tz=timezone.utc,
                ).date()
                if is_published and assignment_time != 0
                else None
            )
        for completed_form in forms.completedForms:
            is_published = self.is_published(completed_form.urn)
            assignment_time = self.assignment_time(completed_form)
            form_assigned_dates[completed_form.urn] = (
                datetime.fromtimestamp(
                    assignment_time,
                    tz=timezone.utc,
                ).date()
                if is_published and assignment_time != 0
                else None
            )
        return form_assigned_dates

    def form_completed_date(
        self, search_row: DataHubDatasetSearchRow
    ) -> Dict[str, date]:
        # get the maximum timestamp from the
        # incompleteFormsCompletedPromptResponseTimes and
        # completedFormsCompletedPromptResponseTimes
        form_completion_dates = {}
        for form in search_row.completedForms:
            form_info = self.form_registry.get_form(form)
            if not form_info:
                logger.warning(f"Found form attached that does not exist: {form}")
                continue
            form_prompts = [x.id for x in form_info.prompts]
            completed_prompts_map = {
                prompt_id: response_time
                for prompt_id, response_time in zip(
                    search_row.completedFormsCompletedPromptIds,
                    search_row.completedFormsCompletedPromptResponseTimes,
                    strict=False,
                )
                if prompt_id in form_prompts
            }
            sorted_dates = sorted(completed_prompts_map.values())
            if sorted_dates:
                form_completion_dates[form] = datetime.fromtimestamp(
                    int(sorted(completed_prompts_map.values())[-1]) / 1000,
                    tz=timezone.utc,
                ).date()
            else:
                logger.warning(
                    f"Form {form} on asset {search_row.urn} has no completed prompts but is marked as completed"
                )
        return form_completion_dates

    def get_assignees(self, form_info: FormInfoClass, owners: List[str]) -> List[str]:
        assignees = []
        actors = form_info.actors
        if actors.owners:
            # form applies to owners
            assignees.extend(owners)
        if actors.users:
            assignees.extend(actors.users)
        if actors.groups:
            assignees.extend(actors.groups)
        return list(set(assignees))

    def get_data(  # noqa: C901
        self,
        on_asset_scanned: Optional[Callable[[str], Any]] = None,
        on_form_scanned: Optional[Callable[[str], Any]] = None,
    ) -> Iterable[FormReportingRow]:
        extra_fields = [f for f in self.DataHubDatasetSearchRow.__fields__]
        # TODO: Replace with the new search/filter SDK.
        result = self.graph.get_results_by_filter(
            extra_or_filters=self.get_form_existence_or_filters(),
            extra_source_fields=extra_fields,
            skip_cache=True,
        )
        forms_scanned = set()
        row_index = 0
        for row in result:
            row_index += 1
            if row_index % 100 == 0:
                logger.info(f"Scanned {row_index} assets")
            extra_properties = row["extraProperties"]

            extra_properties_map = {
                x["name"]: json.loads(x["value"]) for x in extra_properties
            }
            search_row = self.DataHubDatasetSearchRow(**extra_properties_map)
            if on_asset_scanned:
                on_asset_scanned(search_row.urn)
            form_assigned_dates = self.form_assigned_date(search_row)
            form_completed_dates = self.form_completed_date(search_row)
            domain = None
            parent_domain = None
            if search_row.domains:
                domain = search_row.domains[0]
                domain_props = self.domain_registry.get_domain(domain)
                if not domain_props:
                    logger.warning(
                        f"Domain {search_row.domains[0]} not found. Will record the domain urn regardless"
                    )
                else:
                    if domain_props.parentDomain:
                        parent_domain = domain_props.parentDomain
            # for owner in owners:
            incomplete_forms = (
                [x for x in search_row.incompleteForms if x in self.allowed_forms]
                if self.allowed_forms
                else search_row.incompleteForms
            )
            for form_id in incomplete_forms:
                if form_id not in forms_scanned:
                    if on_form_scanned:
                        on_form_scanned(form_id)
                    forms_scanned.add(form_id)
                form_info = self.form_registry.get_form(form_id)
                if not form_info:
                    logger.warning(
                        f"Found form attached that does not exist: {form_id}"
                    )
                    continue
                form_prompts = [x.id for x in form_info.prompts]
                form_incomplete_prompts = [
                    p
                    for p in search_row.incompleteFormsIncompletePromptIds
                    if p in form_prompts
                ]
                form_status = (
                    FormStatus.IN_PROGRESS
                    if len(form_incomplete_prompts) < len(form_prompts)
                    else FormStatus.NOT_STARTED
                )
                form_type = (
                    FormType.DOCUMENTATION
                    if form_info.type == FormTypeClass.COMPLETION
                    else FormType.VERIFICATION
                )
                assignees = self.get_assignees(form_info, search_row.owners)
                for prompt_id in [
                    p
                    for p in search_row.incompleteFormsIncompletePromptIds
                    if p in form_prompts
                ]:
                    for owner in assignees:
                        yield FormReportingRow(
                            form_urn=form_id,
                            form_assigned_date=form_assigned_dates[form_id],
                            form_completed_date=None,
                            form_status=form_status,
                            form_type=form_type,
                            assignee_urn=owner,
                            asset_urn=search_row.urn,
                            platform_urn=search_row.platform,
                            platform_instance_urn=search_row.platformInstance,
                            domain_urn=domain,
                            parent_domain_urn=parent_domain,
                            asset_verified=(
                                False if form_type == FormType.VERIFICATION else None
                            ),
                            question_id=str(prompt_id),
                            question_status=QuestionStatus.NOT_STARTED,
                            question_completed_date=None,
                            snapshot_date=self.snapshot_date,
                        )
                for prompt_id, prompt_response_time in [
                    (p, p_response_time)
                    for (p, p_response_time) in zip(
                        search_row.incompleteFormsCompletedPromptIds,
                        search_row.incompleteFormsCompletedPromptResponseTimes,
                        strict=False,
                    )
                    if p in form_prompts
                ]:
                    for owner in assignees:
                        yield FormReportingRow(
                            form_urn=form_id,
                            form_assigned_date=form_assigned_dates[form_id],
                            form_completed_date=None,
                            form_status=form_status,
                            form_type=form_type,
                            assignee_urn=owner,
                            asset_urn=search_row.urn,
                            platform_urn=search_row.platform,
                            platform_instance_urn=search_row.platformInstance,
                            domain_urn=domain,
                            parent_domain_urn=parent_domain,
                            asset_verified=(
                                False if form_type == FormType.VERIFICATION else None
                            ),
                            question_id=str(prompt_id),
                            question_status=QuestionStatus.COMPLETED,
                            question_completed_date=datetime.fromtimestamp(
                                float(prompt_response_time) / 1000, tz=timezone.utc
                            ),
                            snapshot_date=self.snapshot_date,
                        )
            complete_forms = (
                [x for x in search_row.completedForms if x in self.allowed_forms]
                if self.allowed_forms
                else search_row.completedForms
            )
            verified_forms_list = (
                [x for x in search_row.verifiedForms if x in self.allowed_forms]
                if self.allowed_forms
                else search_row.verifiedForms
            )
            verified_forms = set(verified_forms_list)
            for form_id in complete_forms:
                if form_id not in forms_scanned:
                    if on_form_scanned:
                        on_form_scanned(form_id)
                    forms_scanned.add(form_id)
                form_info = self.form_registry.get_form(form_id)
                if not form_info:
                    logger.warning(
                        f"Found form attached that does not exist: {form_id}"
                    )
                    continue
                form_type = (
                    FormType.DOCUMENTATION
                    if form_info.type == FormTypeClass.COMPLETION
                    else FormType.VERIFICATION
                )
                is_verification_form = form_type == FormType.VERIFICATION
                is_form_verified = form_id in verified_forms
                form_status = (
                    FormStatus.COMPLETED
                    if is_form_verified or not is_verification_form
                    else FormStatus.IN_PROGRESS
                )
                assignees = self.get_assignees(form_info, search_row.owners)
                form_prompts = [x.id for x in form_info.prompts]
                for prompt_id in [
                    p
                    for p in search_row.completedFormsIncompletePromptIds
                    if p in form_prompts
                ]:
                    for owner in assignees:
                        yield FormReportingRow(
                            form_urn=form_id,
                            form_assigned_date=form_assigned_dates[form_id],
                            form_completed_date=form_completed_dates.get(
                                form_id, form_assigned_dates[form_id]
                            ),
                            form_status=form_status,
                            form_type=form_type,
                            assignee_urn=owner,
                            asset_urn=search_row.urn,
                            platform_urn=search_row.platform,
                            platform_instance_urn=search_row.platformInstance,
                            domain_urn=domain,
                            parent_domain_urn=parent_domain,
                            asset_verified=(
                                True
                                if is_verification_form and is_form_verified
                                else None
                            ),
                            question_id=str(prompt_id),
                            question_status=QuestionStatus.NOT_STARTED,
                            question_completed_date=None,
                            snapshot_date=self.snapshot_date,
                        )
                for prompt_id, prompt_response_time in [
                    (p, p_response_time)
                    for (p, p_response_time) in zip(
                        search_row.completedFormsCompletedPromptIds,
                        search_row.completedFormsCompletedPromptResponseTimes,
                        strict=False,
                    )
                    if p in form_prompts
                ]:
                    for owner in assignees:
                        yield FormReportingRow(
                            form_urn=form_id,
                            form_assigned_date=form_assigned_dates[form_id],
                            form_completed_date=form_completed_dates.get(
                                form_id, form_assigned_dates[form_id]
                            ),
                            form_status=form_status,
                            form_type=form_type,
                            assignee_urn=owner,
                            asset_urn=search_row.urn,
                            platform_urn=search_row.platform,
                            platform_instance_urn=search_row.platformInstance,
                            domain_urn=domain,
                            parent_domain_urn=parent_domain,
                            asset_verified=(
                                True
                                if is_verification_form and is_form_verified
                                else None
                            ),
                            question_id=str(prompt_id),
                            question_status=QuestionStatus.COMPLETED,
                            question_completed_date=datetime.fromtimestamp(
                                float(prompt_response_time) / 1000, tz=timezone.utc
                            ),
                            snapshot_date=self.snapshot_date,
                        )


if __name__ == "__main__":
    from datahub.ingestion.graph.client import get_default_graph

    with get_default_graph() as graph:
        form_data = DataHubFormReportingData(graph)
        for row in form_data.get_data():
            print(row)



"""
Shared constants for column metric assertions (both smart and non-smart).

This module contains constants that are used by both smart and non-smart column metric assertions
to ensure consistency and avoid duplication.
"""

from enum import Enum
from typing import Union

from datahub.metadata import schema_classes as models

# Keep this in sync with the frontend in getEligibleFieldColumns
# datahub-web-react/src/app/entityV2/shared/tabs/Dataset/Validations/assertion/builder/steps/field/utils.ts
ALLOWED_COLUMN_TYPES_FOR_COLUMN_METRIC_ASSERTION = [
    models.StringTypeClass(),
    models.NumberTypeClass(),
    models.BooleanTypeClass(),
    models.DateTypeClass(),
    models.TimeTypeClass(),
    models.NullTypeClass(),
]

# Keep this in sync with FIELD_VALUES_OPERATOR_CONFIG in the frontend
# datahub-web-react/src/app/entityV2/shared/tabs/Dataset/Validations/assertion/builder/steps/field/utils.ts
FIELD_VALUES_OPERATOR_CONFIG = {
    "STRING": [
        models.AssertionStdOperatorClass.NULL,
        models.AssertionStdOperatorClass.NOT_NULL,
        models.AssertionStdOperatorClass.EQUAL_TO,
        models.AssertionStdOperatorClass.IN,
        models.AssertionStdOperatorClass.GREATER_THAN_OR_EQUAL_TO,
        models.AssertionStdOperatorClass.REGEX_MATCH,
        models.AssertionStdOperatorClass.GREATER_THAN,
        models.AssertionStdOperatorClass.LESS_THAN,
        models.AssertionStdOperatorClass.BETWEEN,
    ],
    "NUMBER": [
        models.AssertionStdOperatorClass.GREATER_THAN,
        models.AssertionStdOperatorClass.LESS_THAN,
        models.AssertionStdOperatorClass.BETWEEN,
        models.AssertionStdOperatorClass.NULL,
        models.AssertionStdOperatorClass.NOT_NULL,
        models.AssertionStdOperatorClass.EQUAL_TO,
        models.AssertionStdOperatorClass.IN,
        models.AssertionStdOperatorClass.GREATER_THAN_OR_EQUAL_TO,
        models.AssertionStdOperatorClass.NOT_EQUAL_TO,
    ],
    "BOOLEAN": [
        models.AssertionStdOperatorClass.IS_TRUE,
        models.AssertionStdOperatorClass.IS_FALSE,
        models.AssertionStdOperatorClass.NULL,
        models.AssertionStdOperatorClass.NOT_NULL,
    ],
    "DATE": [
        models.AssertionStdOperatorClass.NULL,
        models.AssertionStdOperatorClass.NOT_NULL,
    ],
    "TIME": [
        models.AssertionStdOperatorClass.NULL,
        models.AssertionStdOperatorClass.NOT_NULL,
    ],
    "NULL": [
        models.AssertionStdOperatorClass.NULL,
        models.AssertionStdOperatorClass.NOT_NULL,
    ],
}

# Keep this in sync with FIELD_METRIC_TYPE_CONFIG in the frontend
# datahub-web-react/src/app/entityV2/shared/tabs/Dataset/Validations/assertion/builder/steps/field/utils.ts
FIELD_METRIC_TYPE_CONFIG = {
    "STRING": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
        models.FieldMetricTypeClass.MAX_LENGTH,
        models.FieldMetricTypeClass.MIN_LENGTH,
        models.FieldMetricTypeClass.EMPTY_COUNT,
        models.FieldMetricTypeClass.EMPTY_PERCENTAGE,
    ],
    "NUMBER": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
        models.FieldMetricTypeClass.MAX,
        models.FieldMetricTypeClass.MIN,
        models.FieldMetricTypeClass.MEAN,
        models.FieldMetricTypeClass.MEDIAN,
        models.FieldMetricTypeClass.STDDEV,
        models.FieldMetricTypeClass.NEGATIVE_COUNT,
        models.FieldMetricTypeClass.NEGATIVE_PERCENTAGE,
        models.FieldMetricTypeClass.ZERO_COUNT,
        models.FieldMetricTypeClass.ZERO_PERCENTAGE,
    ],
    "BOOLEAN": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
    ],
    "DATE": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
    ],
    "TIME": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
    ],
    "NULL": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
    ],
}


class MetricType(str, Enum):
    """Enum for field metric types used in column metric assertions."""

    NULL_COUNT = models.FieldMetricTypeClass.NULL_COUNT
    NULL_PERCENTAGE = models.FieldMetricTypeClass.NULL_PERCENTAGE
    UNIQUE_COUNT = models.FieldMetricTypeClass.UNIQUE_COUNT
    UNIQUE_PERCENTAGE = models.FieldMetricTypeClass.UNIQUE_PERCENTAGE
    MAX_LENGTH = models.FieldMetricTypeClass.MAX_LENGTH
    MIN_LENGTH = models.FieldMetricTypeClass.MIN_LENGTH
    EMPTY_COUNT = models.FieldMetricTypeClass.EMPTY_COUNT
    EMPTY_PERCENTAGE = models.FieldMetricTypeClass.EMPTY_PERCENTAGE
    MIN = models.FieldMetricTypeClass.MIN
    MAX = models.FieldMetricTypeClass.MAX
    MEAN = models.FieldMetricTypeClass.MEAN
    MEDIAN = models.FieldMetricTypeClass.MEDIAN
    STDDEV = models.FieldMetricTypeClass.STDDEV
    NEGATIVE_COUNT = models.FieldMetricTypeClass.NEGATIVE_COUNT
    NEGATIVE_PERCENTAGE = models.FieldMetricTypeClass.NEGATIVE_PERCENTAGE
    ZERO_COUNT = models.FieldMetricTypeClass.ZERO_COUNT
    ZERO_PERCENTAGE = models.FieldMetricTypeClass.ZERO_PERCENTAGE


class OperatorType(str, Enum):
    """Enum for assertion operators used in column metric assertions."""

    EQUAL_TO = models.AssertionStdOperatorClass.EQUAL_TO
    NOT_EQUAL_TO = models.AssertionStdOperatorClass.NOT_EQUAL_TO
    GREATER_THAN = models.AssertionStdOperatorClass.GREATER_THAN
    GREATER_THAN_OR_EQUAL_TO = models.AssertionStdOperatorClass.GREATER_THAN_OR_EQUAL_TO
    LESS_THAN = models.AssertionStdOperatorClass.LESS_THAN
    LESS_THAN_OR_EQUAL_TO = models.AssertionStdOperatorClass.LESS_THAN_OR_EQUAL_TO
    BETWEEN = models.AssertionStdOperatorClass.BETWEEN
    IN = models.AssertionStdOperatorClass.IN
    NOT_IN = models.AssertionStdOperatorClass.NOT_IN
    NULL = models.AssertionStdOperatorClass.NULL
    NOT_NULL = models.AssertionStdOperatorClass.NOT_NULL
    IS_TRUE = models.AssertionStdOperatorClass.IS_TRUE
    IS_FALSE = models.AssertionStdOperatorClass.IS_FALSE
    CONTAIN = models.AssertionStdOperatorClass.CONTAIN
    END_WITH = models.AssertionStdOperatorClass.END_WITH
    START_WITH = models.AssertionStdOperatorClass.START_WITH
    REGEX_MATCH = models.AssertionStdOperatorClass.REGEX_MATCH


class ValueType(str, Enum):
    """Enum for assertion parameter value types."""

    STRING = models.AssertionStdParameterTypeClass.STRING
    NUMBER = models.AssertionStdParameterTypeClass.NUMBER
    UNKNOWN = models.AssertionStdParameterTypeClass.UNKNOWN
    # Note: LIST and SET are intentionally excluded as they are not yet supported
    # LIST = models.AssertionStdParameterTypeClass.LIST
    # SET = models.AssertionStdParameterTypeClass.SET


# Type aliases
MetricInputType = Union[MetricType, models.FieldMetricTypeClass, str]
ValueInputType = Union[str, int, float]
ValueTypeInputType = Union[ValueType, models.AssertionStdParameterTypeClass, str]
RangeInputType = tuple[ValueInputType, ValueInputType]
RangeTypeInputType = Union[
    str,
    tuple[str, str],
    ValueTypeInputType,
    tuple[ValueTypeInputType, ValueTypeInputType],
]
RangeTypeParsedType = tuple[ValueTypeInputType, ValueTypeInputType]
OperatorInputType = Union[OperatorType, models.AssertionStdOperatorClass, str]
